#include "gba.h"
#include "graphic.h"
#include "ninja.hpp"
#include "level.hpp"

#include "../graphics/spriteset.h"
#include "../graphics/tileset.h"

#define BUMPER_START 		4
#define BUMPER_COUNT 		16
#define BUMPER_ANIM_START	576
#define BUMPER_ANIM_END		584

// used for diffs to stream the vram
volatile s32 old_scroll_x;
volatile s32 old_scroll_y;

volatile OAM_OBJECT oam_buffer[128];
s8 animated[128];

void init_game_graphics()
{
	// setup BG display with OBJ and force VBlank while setting up
	*REG_DISPCNT = BG1_ENABLE | BG2_ENABLE | OBJ_ENABLE | MODE_0 | FORCE_VBLANK;
	*REG_BG1CNT = CHAR_BASE_0000 | SCREEN_BASE_D800 | COLOR_16 | PRIORITY_1;
	*REG_BG2CNT = CHAR_BASE_0000 | SCREEN_BASE_D000 | COLOR_16;

	// copy tilesets
	*REG_DMA3SAD = tilesetTiles;
	*REG_DMA3DAD = _vram;
	*REG_DMA3CNT = DMA_ENABLE | DMA_IMMEDIATE | DMA_32 | sizeof(tilesetTiles)/sizeof(u32);

	// copy tilesets
	*REG_DMA3SAD = tilesetPal;
	*REG_DMA3DAD = _palette;
	*REG_DMA3CNT = DMA_ENABLE | DMA_IMMEDIATE | DMA_32 | sizeof(tilesetPal)/sizeof(u32);

	*REG_DMA3SAD = spritesetTiles;
	*REG_DMA3DAD = _vram_spr;
	*REG_DMA3CNT = DMA_ENABLE | DMA_IMMEDIATE | DMA_32 | sizeof(spritesetTiles)/sizeof(u32);

	*REG_DMA3SAD = spritesetPal;
	*REG_DMA3DAD = _obj_palette;
	*REG_DMA3CNT = DMA_ENABLE | DMA_IMMEDIATE | DMA_16 | sizeof(spritesetPal)/sizeof(u16);

	for (u16 i = 0; i < 32*20; ++i)
	{
		*(vram(0xD800) + i) = i + 192 | 0x6000;
	}
	*REG_BG1HOFS = 16;
	*REG_BG1VOFS = 0;

	clear_oam();
}

void flush_graphics()
{
	s32 scroll_x = game.scroll_x;
	s32 scroll_y = game.scroll_y;

	*REG_BG1HOFS = scroll_x/2;
	*REG_BG2HOFS = scroll_x;
	*REG_BG2VOFS = scroll_y;

	scroll_x = scroll_x/8;
	scroll_y = scroll_y/8;

	for (s32 x = scroll_x; x < scroll_x+32; ++x)
	{
		for (s32 y = scroll_y; y < scroll_y+32; ++y)
		{
			*(vram(0xD000) + 32*(y % 32) + (x % 32)) = game.level.fasttile(x, y).tile;
		}
	}

	oam_buffer[3].attr2 = 4;
	oam_buffer[2].attr2 = 132;
	oam_buffer[1].attr2 = 260;
	oam_buffer[0].attr2 = 324;

	old_scroll_x = scroll_x;
	old_scroll_y = scroll_y;
}

inline bool check_inview(s32 x, s32 y)
{
	return (x > -128 && x < RESOLUTION_X) && y > -64&& y < RESOLUTION_Y;
}

inline u16 _get_character_animation(u16 old_anim, s8 dir, s8 & status)
{
	if (status < 0)
	{
		--status;
		if (status <= -8)
		{
			status = -1;
			if (old_anim == 0)
			{
				return 4;
			}
			if (old_anim == 5)
			{
				status = 0;
				return 4;
			}
			return 5;
		}

		return old_anim;
	}
	if (old_anim == 4)
	{
		++status;
		if (status >= 7)
		{
			status = 0;
			return 0;
		}
		return 4;
	}

	if (dir == -2 || dir == 2)
	{
		status += 2;
	}
	else
	{
		status += 1;
	}

	if (status >= 30)
	{
		status = 0;
		if (old_anim == 3)
		{
			return 0;
		}

		return old_anim + 1;
	}
	return old_anim;
}

#define get_character_attr2(attr2, base_tile, size, dir, status)\
	_get_character_animation(((attr2 & 0x03ff) - base_tile)/size, dir, status)*size + base_tile


void prepare_draw()
{
	s32 scroll_x = game.scroll_x;
	s32 scroll_y = game.scroll_y;

	s32 x = game.avatars[0].box.pos.x - scroll_x;
	s32 y = game.avatars[0].box.pos.y - scroll_y;

	if (!(game.avatars[0].active && check_inview(x, y)))
	{
		oam_buffer[3].attr0 = OM_HIDE;
	}
	else
	{
		oam_buffer[3].attr0 = (y & 0x00ff);
		oam_buffer[3].attr1 = SIZE_32 | (x & 0x01ff);
		oam_buffer[3].attr2 = ATTR2_PALETTE(1) | get_character_attr2(oam_buffer[3].attr2, 4, 4, game.avatars[0].dir, animated[3]);

		if (game.avatars[0].dir < 0)
		{
			oam_buffer[3].attr1 |= HFLIP;
		}
	}

	x = game.avatars[1].box.pos.x - scroll_x ;
	y = game.avatars[1].box.pos.y - scroll_y;

	if (!(game.avatars[1].active && check_inview(x, y)))
	{
		oam_buffer[2].attr0 = OM_HIDE;
	}
	else
	{
		oam_buffer[2].attr0 = (y & 0x00ff);
		oam_buffer[2].attr2 = ATTR2_PALETTE(2) | get_character_attr2(oam_buffer[2].attr2, 132, 4, game.avatars[1].dir, animated[2]);

		if (game.avatars[1].dir == -1)
		{
			oam_buffer[2].attr1 = SIZE_32 | HFLIP | ((x - 8) & 0x01ff);
		}
		else
		{
			oam_buffer[2].attr1 = SIZE_32 | (x & 0x01ff);
		}
	}

	x = game.avatars[2].box.pos.x - scroll_x;
	y = game.avatars[2].box.pos.y - scroll_y;

	if (!(game.avatars[2].active && check_inview(x, y)))
	{
		oam_buffer[1].attr0 = OM_HIDE;
	}
	else
	{
		oam_buffer[1].attr0 = y & 0x00ff;
		oam_buffer[1].attr1 = SIZE_16 | (x & 0x01ff);
		oam_buffer[1].attr2 = ATTR2_PALETTE(3) | get_character_attr2(oam_buffer[1].attr2, 260, 2, game.avatars[1].dir, animated[1]);;

		if (game.avatars[2].dir == -1)
		{
			oam_buffer[1].attr1 |= HFLIP;
		}
	}

	x = game.avatars[3].box.pos.x - scroll_x;
	y = game.avatars[3].box.pos.y - scroll_y;

	if (!(game.avatars[3].active && check_inview(x, y)))
	{
		oam_buffer[0].attr0 = OM_HIDE;
	}
	else
	{
		oam_buffer[0].attr0 = y & 0x00ff;
		oam_buffer[0].attr1 = SIZE_8  | (x & 0x01ff);
		oam_buffer[0].attr2 = ATTR2_PALETTE(4) | get_character_attr2(oam_buffer[0].attr2, 324, 1, game.avatars[1].dir, animated[0]);;

		if (game.avatars[3].dir == -1)
		{
			oam_buffer[0].attr1 |= HFLIP;
		}
	}

	for (u8 i = 0; i < BUMPER_COUNT; ++i)
	{
		oam_buffer[i + BUMPER_START].attr0 = OM_HIDE;
	}

	for (u8 i = 0; i < game.level.bumperCount(); ++i)
	{
		x = game.level.bumper(i).x - scroll_x;
		y = game.level.bumper(i).y - scroll_y;
		if (check_inview(x, y))
		{
			u8 index = (i % 32) + BUMPER_START;
			oam_buffer[index].attr0 = y & 0x00ff;
			oam_buffer[index].attr1 = SIZE_32 | (x & 0x01ff);
			oam_buffer[index].attr2 = ATTR2_PALETTE(2) | BUMPER_ANIM_START;

			if (animated[index] > 0)
			{
				++animated[index];
				if (animated[index] < 12)
				{
					oam_buffer[index].attr2 += 4*(animated[index]/4);
				}
				else
				{
					animated[index] = 0;
				}
			}
		}
	}

	for (u8 i = 0; i < BUMPER_COUNT; ++i)
	{
		u8 index = i + BUMPER_START;
		if (oam_buffer[index].attr0 == OM_HIDE)
		{
			animated[index] = 0;
		}
	}
}

void draw()
{
	// copy the OAM buffer
	*REG_DMA3SAD = (u16*)oam_buffer;
	*REG_DMA3DAD = _oam;
	*REG_DMA3CNT = DMA_ENABLE | DMA_IMMEDIATE | DMA_32 | sizeof(oam_buffer)/sizeof(u32);

	s32 scroll_x = game.scroll_x;
	s32 scroll_y = game.scroll_y;

	*REG_BG1HOFS = scroll_x/2;
	*REG_BG2HOFS = scroll_x;
	*REG_BG2VOFS = scroll_y;

	scroll_x = scroll_x/8;
	scroll_y = scroll_y/8;

	s32 right = scroll_x + 31;
	s32 old_right = old_scroll_x + 31;

	s32 bot = scroll_y + 21;
	s32 old_bot = old_scroll_y + 21;

	s32 l;
	s32 r;

	s32 t;
	s32 b;

	if (scroll_x < old_scroll_x)
	{
		l = scroll_x;
		r = right < old_scroll_x ? right : old_scroll_x;
	}
	else
	{
		l = old_right > scroll_x ? old_right : scroll_x;
		r = right;
	}

	for (s32 x = l; x < r; ++x)
	{
		for (s32 y = scroll_y; y < bot; ++y)
		{
			*(vram(0xD000) + 32*(y % 32) + (x % 32)) = game.level.r_tiles[x + game.level.m_width*y].tile;
		}
	}

	if (scroll_y < old_scroll_y)
	{
		t = scroll_y;
		b = bot < old_scroll_y ? bot : old_scroll_y;
	}
	else
	{
		t = old_bot > scroll_y ? old_bot : scroll_y;
		b = bot;
	}

	for (s32 x = scroll_x; x < right; ++x)
	{
		for (s32 y = t; y < b; ++y)
		{
			*(vram(0xD000) + 32*(y % 32) + (x % 32)) = game.level.r_tiles[x + game.level.m_width*y].tile;
		}
	}

	old_scroll_x = scroll_x;
	old_scroll_y = scroll_y;
}

void split_avatar(int avatar)
{
	animated[3 - avatar] = -1;
}

void animate_bumper(int bumper)
{
	animated[BUMPER_START + bumper] = 2;
}

void clear_oam()
{
	for (int i = 0; i < 128; ++i)
	{
		oam[i].attr0 = OM_HIDE;
		oam_buffer[i].attr0 = OM_HIDE;
		animated[i] = false;
	}
}
