#ifndef LEVEL_HPP_INCLUDED
#define LEVEL_HPP_INCLUDED

#include "gba.h"
#include "util.hpp"

#define BUMPER_WIDTH_TILES 3
#define BUMPER_HEIGHT_TILES 2

#define TILE_SIZE_PX 8

////////////////////////////////////////////////////////////////////////////////

struct Tile
{
	enum TypeMasks
	{
		EMPTY = 0,
		BLOCK = 1, // First bit to 1 means solid
		LETHAL = 1 << 1,
		CHECKPOINT = 1 << 2,
		COLOR1 = 1 << 3,
		COLOR2 = 2 << 3,
		COLOR3 = 4 << 3,
		COLOR4 = 8 << 3
	};

	// Bits :
	// 1 : solid
	// 2 : lethal
	// 3 : checkpoint
	// 4-7 : color (0000, 0001, 0010, 0100, 1000)
	// 8 : unusued
	u8 type;

	u16 tile; // Display

	inline bool isSolid() const
	{
		return type & 1;
	}

	inline bool isLethal() const
	{
		return type & 2;
	}

	inline bool isCheckPoint() const
	{
		return type & 4;
	}

	//Tile() : type(EMPTY), tile(0) {}
};

inline s32 tilesToPixels(s32 x)
{
	return x * TILE_SIZE_PX;
}

inline s32 pixelsToTiles(s32 x)
{
	return x / TILE_SIZE_PX;
}

////////////////////////////////////////////////////////////////////////////////

class Level
{
#ifdef SENTINELS
private: char __instanceBegin[4];
#endif

public:

	Tile m_outerworldTile;
	const Tile * r_tiles; // [y][x]
	u32 m_width;
	u32 m_height; // I know this will not change, BUT I assume it still can.
	const RawVector2i * r_bumpers;
	u32 m_bumperCount;

	Vector2i spawn;

	Level();
	~Level();

	void load(
		const Tile * tiles, u32 widthTiles, u32 heightTiles,
		const RawVector2i * bumpers, u32 bumperCount);

	inline u32 width() const { return m_width; }
	inline u32 height() const { return m_height; }

	inline bool isLoaded()
	{
		return r_tiles == 0;
	}

	inline bool contains(s32 tx, s32 ty)
	{
		return tx >= 0 && tx < static_cast<s32>(m_width)
			&& ty >= 0 && ty < static_cast<s32>(m_height);
	}

	inline const Tile & tile(const Vector2i & pos)
	{
		return tile(pos.x, pos.y);
	}

	// Accesses a tile. Returns a default tile if positions are invalid.
	inline const Tile & tile(s32 tx, s32 ty)
	{
		if(contains(tx, ty))
			return r_tiles[ty * m_width + tx];
		return m_outerworldTile;
	}

	// Accesses a tile. Returns a default tile if positions are invalid.
	inline const Tile & fasttile(s32 tx, s32 ty)
	{
		return r_tiles[ty * m_width + tx];
	}

	inline const u32 bumperCount() const { return m_bumperCount; }
	inline const RawVector2i & bumper(u32 i) const { return r_bumpers[i]; }

	bool isBumperAtPixels(s32 x, s32 y);
	bool isBumperAtPixels(const Rect & r, u32 * index=0);

	bool pixelCollision(s32 x, s32 y, u8 mask=Tile::BLOCK);

	// Only tests tiles and static objects.
	// If the test matches the mask, the first matching tile position is retrieved
	// to the given tilePos if not null.
	bool pixelCollision(const Rect & r, u8 mask=Tile::BLOCK, Vector2i * tilePos=0);

#ifdef SENTINELS
private: char __instanceEnd[4];
#endif
};


#endif // LEVEL_HPP_INCLUDED



